from lunespy.tx.transfer import TransferToken
from requests import Response

from lunespy.crypto import bytes_to_b58


def serialize_transfer(tx: TransferToken) -> bytes:
    from lunespy.crypto import b58_to_bytes
    from struct import pack

    return (
        chr(tx.type).encode() + \
        b58_to_bytes(tx.senderPublicKey) + \
        (b'\1' + b58_to_bytes(tx.assetId) if tx.assetId != "" else b'\0') + \
        (b'\1' + b58_to_bytes(tx.feeAsset) if tx.feeAsset != "" else b'\0') + \
        pack(">Q", tx.timestamp) + \
        pack(">Q", tx.amount) + \
        pack(">Q", tx.fee) + \
        b58_to_bytes(tx.recipient)
    )


def sign_transfer(private_key: str, tx: TransferToken) -> dict:
    from lunespy.crypto import fast_signature
    from lunespy.crypto import b58_to_bytes, bytes_to_b58

    tx.message = bytes_to_b58(serialize_transfer(tx))

    return bytes_to_b58(fast_signature(
        b58_to_bytes(private_key),
        b58_to_bytes(tx.message)
    ))



# todo async
def broadcast_transfer(mount_tx: dict, node_url: str) -> Response:
    from httpx import post

    return post(
        f'{node_url}/transactions/broadcast',
        json=mount_tx,
        headers={
            'content-type':
            'application/json'
    })



def validate_signature(public_key: bytes, message: bytes, signature: bytes) -> bool:
    from axolotl_curve25519 import verifySignature

    return True if verifySignature(
        public_key,
        message,
        signature
    ) == 0 else False

def fast_signature(private_key: bytes, message: bytes) -> bytes:
    from axolotl_curve25519 import calculateSignature as curve
    from os import urandom

    return curve(
        urandom(64),
        private_key,
        message
    )

def to_sha256(message: bytes) -> bytes:
    from hashlib import sha256

    return sha256(message).digest()

def to_keccak256(message: bytes) -> bytes:
    from sha3 import keccak_256

    return keccak_256(message).digest()

def to_blake2b32b(message: bytes) -> bytes:
    from hashlib import blake2b

    return blake2b(message, digest_size=32).digest()

def bytes_to_b58(message: bytes) -> str:
    from base58 import b58encode

    return b58encode(message).decode()

def b58_to_bytes(message: str) -> bytes:
    from base58 import b58decode

    return b58decode(message)

def to_address(public_key: bytes, chain: int, addr_version: int) -> bytes:
    raw_addr: bytes = (
        chr(addr_version).encode('latin-1') +
        str(chain).encode('latin-1') +
        to_keccak256(to_blake2b32b(public_key))[:20]
    )
    checksum: bytes = to_keccak256(to_blake2b32b(raw_addr))[:4]

    return (raw_addr + checksum)

def to_private_key(seed: bytes) -> bytes:
    from axolotl_curve25519 import generatePrivateKey

    return generatePrivateKey(seed)

def to_public_key(private_key: bytes) -> bytes:
    from axolotl_curve25519 import generatePublicKey

    return generatePublicKey(private_key)

def hidden_seed(nonce: int, seed: str) -> bytes:
    from struct import pack

    raw_seed: bytes = (pack(">L", nonce) + seed.encode())
    return to_sha256(to_keccak256(to_blake2b32b(raw_seed)))

def validate_address(chain: int, address: str) -> bool:
    from lunespy.wallet.constants import ADDRESS_CHECKSUM_LENGTH, ADDRESS_VERSION, ADDRESS_LENGTH

    raw_address: bytes = b58_to_bytes(address)
    address_left, checksum = raw_address[:-ADDRESS_CHECKSUM_LENGTH], raw_address[-ADDRESS_CHECKSUM_LENGTH:]
    hash_address_left: str = to_keccak256(to_blake2b32b(address_left))
    addr_checksum: str = hash_address_left[:ADDRESS_CHECKSUM_LENGTH]

    if checksum != addr_checksum:
        return False
    elif raw_address[0] not in ADDRESS_VERSION:
        return False
    elif raw_address[1] != str(chain).encode()[0]:
        return False
    elif len(raw_address) != ADDRESS_LENGTH:
        return False

    return True

def same_chain_address(address_1: str, address_2: str) -> bool:
    x = validate_address(1, address_1) and validate_address(1, address_2)
    y = validate_address(0, address_1) and validate_address(0, address_2)

    return x or y


from pydantic import BaseModel, Field


class Wallet(BaseModel):
    nonce: int = Field(0, ge=0, le=4_294_967_296)
    seed_len: int = Field(12, multiple_of=3)
    chain: int = Field(1, le=1, ge=0)
    private_key: str = Field("")
    public_key: str = Field("")
    address: str = Field("")
    seed: str = Field("")


def wallet_factory(
    private_key: str = None,
    seed_len: int = 12,
    seed: str = None,
    chain: int = 1,
    nonce: int = 0
    ) -> Wallet:
    """
        chain: 1 for mainnet, 0 for testnet
    """
    from lunespy.wallet.assembly import from_private_key, from_seed, random_seed

    if private_key == None:
        return from_seed(
            seed if not seed == None else random_seed(seed_len),
            nonce,
            chain
        )
    else:
        return from_private_key(private_key, chain)




from pydantic import BaseModel, Field, validator

class TransferToken(BaseModel):
    from lunespy.utils import now
    from requests import Response

    senderPublicKey: str = Field(..., description="public_key")
    assetId: str = Field("", description="token id | NFT id")
    sender: str = Field("", description="address of sender")
    recipient: str = Field(..., description="address")
    timestamp: int = Field(now(), ge=1483228800)
    message: str = Field("", exclude=True)
    fee: int = Field(1000000, ge=1000000)
    type: int = Field(4, const=True)
    amount: int = Field(..., gt=0)
    signature: str = Field("")
    feeAsset: int = Field("")

    @validator("recipient")
    def same_chain(cls, recipient, values):
        from lunespy.crypto import same_chain_address

        if not same_chain_address(recipient, values['sender']):
            raise ValueError(f"Different chain addresses ({recipient}, {values['sender']})")
        return recipient

    def sign(cls, private_key: str):
        from lunespy.tx.transfer.utils import sign_transfer
        cls.signature = sign_transfer(private_key, cls)
        return cls

    def broadcast(cls, node: str = None) -> Response:
        from lunespy.tx.transfer.utils import broadcast_transfer

        return broadcast_transfer(
            cls.dict(),
            node if not node == None else "https://lunesnode-testnet.lunes.io"
        )


def transfer_token_factory(sender: str, receiver: str, amount: float, chain: int = 1, **kwargs: dict) -> TransferToken:
    from lunespy.crypto import b58_to_bytes, bytes_to_b58, to_address


    return TransferToken(
        sender=bytes_to_b58(to_address(b58_to_bytes(sender), chain, 1)),
        amount=int(amount * 10e7),
        senderPublicKey=sender,
        recipient=receiver,
        **kwargs
    )




def random_seed(len: int) -> str:
    from lunespy.wallet.constants import word_list
    from os import urandom

    def f():
        word_count = 2048
        r: bytes = urandom(4)
        x: int = r[3] + (r[2] << 8) + (r[1] << 16) + (r[0] << 24)
        w1: int = x % word_count
        w2: int = ((int(x / word_count) >> 0) + w1) % word_count
        w3: int = ((int((int(x / word_count) >> 0) / word_count) >> 0) + w2) % word_count
        return w1, w2, w3

    n_words_multiple_of_3: int = len // 3

    return " ".join([
        word_list[n]
        for _ in range(n_words_multiple_of_3)
        for n in f()
    ])


def from_seed(seed: str, nonce: int, chain: int):
    from lunespy.crypto import bytes_to_b58, to_address, to_private_key, to_public_key, hidden_seed
    from lunespy.wallet import Wallet

    hash_seed: bytes = hidden_seed(nonce, seed)
    private_key: bytes = to_private_key(hash_seed)
    public_key: bytes = to_public_key(private_key)
    address: bytes = to_address(public_key, chain, 1)

    return Wallet(
        private_key=bytes_to_b58(private_key),
        public_key=bytes_to_b58(public_key),
        address=bytes_to_b58(address),
        nonce=nonce,
        chain=chain,
        seed=seed
    )


def from_private_key(private_key: str, chain: int = 1):
    from lunespy.crypto import bytes_to_b58, b58_to_bytes, to_address, to_private_key, to_public_key
    from lunespy.wallet import Wallet

    private_key: bytes = to_private_key(b58_to_bytes(private_key))
    public_key: bytes = to_public_key(private_key)
    address: bytes = to_address(public_key, chain, 1)

    return Wallet(
        private_key=bytes_to_b58(private_key),
        public_key=bytes_to_b58(public_key),
        address=bytes_to_b58(address),
        seed_len=0,
        nonce=0,
        chain=chain,
        seed=""
    )




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lunespy',
 'lunespy.client.transactions',
 'lunespy.client.transactions.alias',
 'lunespy.client.transactions.burn',
 'lunespy.client.transactions.cancel_lease',
 'lunespy.client.transactions.issue',
 'lunespy.client.transactions.lease',
 'lunespy.client.transactions.mass',
 'lunespy.client.transactions.reissue',
 'lunespy.crypto',
 'lunespy.server.address',
 'lunespy.server.blocks',
 'lunespy.server.nodes',
 'lunespy.server.transactions',
 'lunespy.tx.transfer',
 'lunespy.utils',
 'lunespy.wallet']

package_data = \
{'': ['*']}

install_requires = \
['base58>=2.1.0,<3.0.0',
 'httpx>=0.22.0,<0.23.0',
 'pydantic>=1.9.0,<2.0.0',
 'pysha3>=1.0.2,<2.0.0',
 'python-axolotl-curve25519>=0.4.1.post2,<0.5.0',
 'requests>=2.26.0,<3.0.0']

setup_kwargs = {
    'name': 'lunespy',
    'version': '2.0.0',
    'description': 'Library for communication with nodes in mainnet or testnet of the lunes-blockchain network Allows the automation of sending assets, issue end reissue tokens, leasing, registry, and create new wallet.',
    'long_description': '<p align="center">\n    <img alt="Lunes" src="docs/logo.png" width="100" />\n</p>\n\n# LunesPy\n\n- Library for communication with nodes in mainnet or testnet of the lunes-blockchain network\n  Allows the automation of **sending assets**, **issue end reissue tokens**, **lease** and **create new wallet**.\n\n[![Test](https://github.com/lunes-platform/lunespy/actions/workflows/python-app.yml/badge.svg)](https://github.com/lunes-platform/lunespy/actions/workflows/python-app.yml)\n[![License](https://img.shields.io/github/license/lunes-platform/lunespy?color=blueviolet)](LICENSE)\n[![Stars](https://img.shields.io/github/stars/lunes-platform/lunespy?color=blueviolet)](https://github.com/lunes-platform/lunespy/stargazers)\n\n[![CommitActivity](https://img.shields.io/github/commit-activity/m/lunes-platform/lunespy?color=blueviolet)](https://github.com/lunes-platform/lunespy/pulse)\n[![Forks](https://img.shields.io/github/forks/lunes-platform/lunespy?color=blueviolet)](https://github.com/lunes-platform/lunespy/network/members)\n[![Contributors](https://flat.badgen.net/github/contributors/lunes-platform/lunespy?color=purple)](https://github.com/lunes-platform/lunespy/graphs/contributors)\n![ClosedIssue](https://flat.badgen.net/github/closed-issues/lunes-platform/lunespy?color=red)\n[![Branches](https://badgen.net/github/branches/lunes-platform/lunespy?color=blueviolet)](https://github.com/lunes-platform/lunespy/branches)\n[![Watchers](https://img.shields.io/github/watchers/lunes-platform/lunespy.svg?color=blueviolet)](https://github.com/lunes-platform/lunespy/watchers)\n[![Followers](https://img.shields.io/github/followers/lunes-platform.svg?style=social&label=Follow&maxAge=2592000?color=blueviolet)](https://github.com/lunes-platform?tab=followers)\n[![Website](https://img.shields.io/website?url=https%3A%2F%2Flunes.io%2F)](https://lunes.io)\n![PullRequest](https://img.shields.io/github/issues-pr/lunes-platform/lunespy?color=blueviolet)\n![PullRequestClosed](https://img.shields.io/github/issues-pr-closed/lunes-platform/lunespy?color=blueviolet)\n\n<br><br>\n<a href="https://twitter.com/LunesPlatform" target="_blank">\n<img alt="Twitter: Lunes Platform" src="https://badgen.net/twitter/follow/lunesplatform?icon=twitter&label=follow @LunesPlatform&color=blue" />\n</a>  \n <a href="https://t.me/LunesPlatformPT" target="_blank">\n<img alt="Telegram" src="https://badgen.net/badge/icon/Lunes%20Platform?icon=telegram&label=Telegram&color=blue"/>\n</a>\n\n## Documentation\n\nThe `lunespy` documentations is hosted at [Telescope](https://blockchain.lunes.io/telescope/)\n\n## Changelog\n\nThe changelog process for this project is described [here](CHANGELOG.md).\n\n## Contributing\n\n`lunespy` is still under development. Contributions are always welcome! Please follow the [Developers Guide](CONTRIBUTING.md) if you want to help.\n\nThanks to the following people who have contributed to this project:\n\n- [olivmath](https://github.com/olivmath)\n- [marcoslkz](https://github.com/marcoslkz)\n- [VanJustin](https://github.com/VanJustin)\n- [xonfps](https://github.com/xonfps)\n\n## Contact\n\nIf you want to contact me you can reach me at <development@lunes.io>.\n\n## License\n\n[Apache License Version 2.0](https://github.com/lunes-platform/lunespy/blob/main/LICENSE).\n',
    'author': 'Lunes Platform',
    'author_email': 'development@lunes.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0',
}


setup(**setup_kwargs)
